/**
 * NEWS Project
 *
 * File:         AnomalyDetector.java
 * RCS:          $Id: AnomalyDetector.java,v 1.35 2009/11/29 19:21:54 drc915 Exp $
 * Description:  AnomalyDetector class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 7, 2008 at 9:47:57 AM
 * Language:     Java
 * Package:      edu.northwestern.news.anomaly
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.news.anomaly;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;
import java.util.Map.Entry;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.logging.LoggerChannel;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.utils.LocaleUtilities;

import com.azureus.plugins.aznetmon.main.RSTPacketStats;

import edu.northwestern.news.MainGeneric;
import edu.northwestern.news.NEWSConfiguration;
import edu.northwestern.news.anomaly.AnomalyProperties.MovingAverageConfig;
import edu.northwestern.news.anomaly.AnomalyProperties.MovingAverageHistory;
import edu.northwestern.news.asn.NetworkAdminASNImpl;
import edu.northwestern.news.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.news.dht.IDDBReadAction;
import edu.northwestern.news.dht.IDDBWriteAction;
import edu.northwestern.news.dht.IDHTOption;
import edu.northwestern.news.dht.IDistributedDatabase;
import edu.northwestern.news.dht.IDistributedDatabaseEvent;
import edu.northwestern.news.dht.azureus.IDHTWriteAction;
import edu.northwestern.news.stats.Statistics;
import edu.northwestern.news.timer.ITimer;
import edu.northwestern.news.timer.ITimerEvent;
import edu.northwestern.news.timer.ITimerEventPerformer;
import edu.northwestern.news.util.PingManager;
import edu.northwestern.news.util.PluginInterface;
import edu.northwestern.news.util.TraceRouteRunner;
import edu.northwestern.news.util.Util;
import edu.northwestern.news.Main;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The AnomalyDetector class uses various statistics and semantic 
 * information to determine if an anomaly has occurred locally. 
 * 
 * Major components:
 * 1) Calculate moving averages for various data sources
 * 2) If above threshold, publish to DHT
 * 3) If other anomalies in DHT, raise alarm
 */
public class AnomalyDetector implements ITimerEventPerformer {
	
	public class AnomalyWriteAction implements IDDBWriteAction{

//		@Override
		public void handleComplete(IDistributedDatabaseEvent event) {
			Main.getLogger().log("Write for "+event.getKey().getKey()+" complete!");
			
		}

//		@Override
		public void handleTimeout(IDistributedDatabaseEvent event) {
			Main.getLogger().log("Write for "+event.getKey().getKey()+" timed out!");
			
		}

//		@Override
		public void handleWrite(IDistributedDatabaseEvent event) {
//			Main.getLogger().log("Anomaly write for "+event.getKey().getKey()+" handled!");
			
		}
		
	}

	

	
	public class AnomalyReadAction implements IDDBReadAction{

//		@Override
		public void handleComplete(IDistributedDatabaseEvent event) {
//			Main.getLogger().log("Read complete!");
			
		}

//		@Override
		public void handleRead(byte[] b, IDistributedDatabaseEvent event) {
			boolean isNoop = false;
			ByteBuffer bBuffer = ByteBuffer.wrap(b);
			bBuffer.order(ByteOrder.LITTLE_ENDIAN);
			AnomalyRecord ar = new AnomalyRecord();
			ar.deserialize(bBuffer);
			
			String key = (String)event.getKey().getKey();
			
			if (Util.currentNEWSTime()-ar.timeGMT>MAX_HISTORY_TIME){
				if (DEBUG) Main.getLogger().log("Reading stale anomaly!");
				return;
			} 
			if (AnomalyRecord.getMyId()==ar.id){
				if (DEBUG) Main.getLogger().log("Reading my own value...");
				return;
			}  else if (ar.id==-1){
				if (DEBUG) Main.getLogger().log("No id  for anomaly record, ignoring!");
				return ;
			} 
			if (!ar.valid){
				return;
			}
			// handle case of "NOOP" report
			if (ar.anomalyTypes==0){
				isNoop = true;
			}
			
			// find all settings where this guy matches
			ArrayList<MovingAverageHistory> matches = new ArrayList<MovingAverageHistory>();
			for (MovingAverageHistory mah : perSettingDetectionData.keySet()){
				if (!TESTING ){
					
					// otherwise look for same settings
					if (ar.windowSize==mah.windowSize 
							&& ar.stdDevThresh <= mah.stdDevThresh
							&& mah.networkName.equals(key))
						matches.add(mah);
				}
			}
			if (matches.size()==0) return;
			
//			if (DEBUG) 
				Main.getLogger().log("Found anomaly from id " +ar.id + "(my id:"+AnomalyRecord.getMyId()+")");
				long anomalyBucket = ar.timeGMT - ar.timeGMT%bucketSize;
			
				// iterate over each match
				for (MovingAverageHistory mah : matches){
					DetectionDataPerSetting ddps = perSettingDetectionData.get(mah);
					
					// add to set of seen independent probabilities
					if (ddps.globalIndepProbs.get(key)==null) 
						ddps.globalIndepProbs.put(key, new HashMap<Integer, Double>());
					// TODO account for time and use latest probs only
					ddps.globalIndepProbs.get(key).put(ar.id, (double)ar.detectionProb);
					if (isNoop)
						continue; // don't do event corroboration for NOOPs					
					
					// also add to set of seen events
					if (ddps.globalAnomalyCounts.get(key)==null) 
						ddps.globalAnomalyCounts.put(key, new HashMap<Long, HashMap<Integer, AnomalyRecord>>());
					if (ddps.globalAnomalyCounts.get(key).get(anomalyBucket)==null)  
						ddps.globalAnomalyCounts.get(key).put(anomalyBucket, new HashMap<Integer, AnomalyRecord>());
					ddps.globalAnomalyCounts.get(key).get(anomalyBucket).put(ar.id, ar);				
					
					// get map of <Region id, text description> for alarms
					HashMap<String, Integer> foundAnomalyTypes = new HashMap<String, Integer>();
					HashMap<String, String> alarms = checkAlarmsV2(foundAnomalyTypes, anomalyBucket, key, ddps);

					// at least one anomalous region
					if (alarms.size()>0 && 
							(ddps.foundAnomalies.get(key)==null || !ddps.foundAnomalies.get(key).contains(anomalyBucket))){
						LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();
						String logString = "";
						String intro = loc.getLocalisedMessageText("news.messages.alarm");
						logString += "["
							+ DateFormat.getDateTimeInstance().format(
									new Date(Util.fromNEWSTime(anomalyBucket))) + "] "+
									intro+"\n";

						for (Entry<String, String> ent : alarms.entrySet()) {
							String parts[] = ent.getKey().split("\\:");
							String region = "";
							if (parts.length==2) region = loc.getLocalisedMessageText("news.regions."+parts[0].toLowerCase()) + 
							": "+parts[1];
							else region = ent.getKey();
							logString+= loc.getLocalisedMessageText("news.misc.region")
							+ " [" + region+"], "+
							//					loc.getLocalisedMessageText("news.misc.anomalous")+ " " +
							//					loc.getLocalisedMessageText("news.misc.sensors")+" "+
							ent.getValue()+"\n";	
							int ratio = Integer.parseInt(ent.getValue().split(" ")[0]);
							// notify listeners
							for (AnomalyListener al : listeners) 
								al.alarmRaised(region, new AnomalyRecord(ar), 
										ddps.globalAnomalyCounts.get(key).get(anomalyBucket).size(),
										anomalyBucket, ratio);
						}
						Main.getLogger().logAlert(LoggerChannel.LT_INFORMATION, 
								logString);

						if (!INSERT_TEST_ANOMALIES)
							Statistics.getInstance().addAnomalyAlarm(ddps.globalAnomalyCounts.get(key).size(), ar, 
									key, foundAnomalyTypes.get(key), anomalyBucket, ar.detectionProb);
						// notify listeners
						//				for (AnomalyListener al : listeners) 
						//					al.alarmRaised(key, ar, 
						//							ddps.globalAnomalyCounts.get(key).size(), 
						//							anomalyBucket, maxRatio);

						ddps.globalAnomalyCounts.get(key).get(anomalyBucket).clear();
						if (ddps.foundAnomalies.get(key)==null) ddps.foundAnomalies.put(key, new HashSet<Long>());
						ddps.foundAnomalies.get(key).add(anomalyBucket);

					} // end if new anomaly found
				} // end for each matching setting
		}

/*		private HashMap<String, String> checkAlarms(HashMap<String, Integer> anomalyTypes) {
			
			// TODO the peers in the larger regions should be exclusive of those in narrower ones
			LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();
			HashMap<String, String> toReturn = new HashMap<String, String>();
			String prefixAlarm="", asAlarm="", countryAlarm = "";
			for (Entry<String, HashMap<Integer, AnomalyRecord>> ent : 
					globalAnomalyCounts.entrySet()){
				int sizeFactor = 1; // prefix
				if (ent.getKey().startsWith("Country")) sizeFactor = 3; // country
				else if (ent.getKey().startsWith("AS")) sizeFactor = 2; // as
				
				HashMap<String, Integer> metricCounts = new HashMap<String, Integer>();
				for (String s : metricIds.keySet()) metricCounts.put(s, 0);
				// figure out if they overlap
				for (AnomalyRecord ar : ent.getValue().values()){
					for (Entry<String, Integer> ent2 : metricIds.entrySet()){
						if ((ar.anomalyTypes&ent2.getValue().intValue())>0){
							metricCounts.put(ent2.getKey(), metricCounts.get(ent2.getKey())+1);																			}
					}
				}
				
				ArrayList<String> alarmMetrics = new ArrayList<String>();
				for (Entry<String, Integer> ent2 : metricCounts.entrySet()){
					if (ent2.getValue()>alarmThreshold*sizeFactor){
						alarmMetrics.add(
								loc.getLocalisedMessageText("news.sensors."+ent2.getKey())+
								" ("+ent2.getValue()+" "+loc.getLocalisedMessageText("news.misc.cases")+")");
						if (anomalyTypes.get(ent.getKey())==null){
							anomalyTypes.put(ent.getKey(), 0);
						}
						anomalyTypes.put(ent.getKey(), anomalyTypes.get(ent.getKey())+metricIds.get(ent2.getKey()));

					}
				}
				// TODO test
				// need more than one anomalous signal!
				if (alarmMetrics.size()>1){
					if (sizeFactor==1) prefixAlarm=ent.getKey();
					else if (sizeFactor==2) asAlarm = ent.getKey();
					else countryAlarm = ent.getKey();
					
					toReturn.put(ent.getKey(), alarmMetrics.toString());
				}
			} // end for each grouping
			
			if (prefixAlarm.length()!=0){
				if (prefixAlarmRaised){
					// suppress if alarm was already raised
					toReturn.remove(prefixAlarm);
					prefixAlarm = "";
					Main.getLogger().log("Suppressing duplicate alarm on prefix...");
				}
				prefixAlarmRaised = true;
			}
			// don't show alarm for AS if prefix isn't affected...
			if (asAlarm.length()>0 ) {
				if (!prefixAlarmRaised){
					toReturn.remove(asAlarm);
					asAlarm = "";
				} else{
					if (asAlarmRaised){
						// suppress if alarm was already raised
						toReturn.remove(asAlarm);
						asAlarm = "";
						Main.getLogger().log("Suppressing duplicate alarm on AS...");
					}
					asAlarmRaised = true;
				}
			}
			// don't show alarm for country if AS isn't affected
			if (countryAlarm.length()>0 ){
				if (!asAlarmRaised){
					toReturn.remove(countryAlarm);
					countryAlarm = "";
				} else{
					if (countryAlarmRaised){
						// suppress if alarm was already raised
						toReturn.remove(countryAlarm);
						countryAlarm = "";
						Main.getLogger().log("Suppressing duplicate alarm on country...");
					}
					countryAlarmRaised = true;
				}
			}
			
			return toReturn;
		}*/
		
		private HashMap<String, String> checkAlarmsV2(HashMap<String, Integer> anomalyTypes, 
				long anomalyBucket, String key, DetectionDataPerSetting ddps) {
			
			// TODO the peers in the larger regions should be exclusive of those in narrower ones
			LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();
			HashMap<String, String> toReturn = new HashMap<String, String>();
			String prefixAlarm="", asAlarm="", countryAlarm = "";
			
			HashMap<Integer, AnomalyRecord> anomalies = 
				ddps.globalAnomalyCounts.get(key).get(anomalyBucket);
				
				if (key.startsWith("Country")){
					// too coarse, not using
				}
				else if (key.startsWith("AS")){
					// at this level, just raise the issue if more than one prefix  
					// in the same AS has a problem
					HashSet<String> prefixes = new HashSet<String>();
					for (Entry<Integer, AnomalyRecord> ent2: anomalies.entrySet()){
						// make sure it is likely to be a network problem
						if (ent2.getValue().detectionProb>alarmThreshold)
							prefixes.add(Util.convertIntToIp(ent2.getKey()));
					}
					// if more than one, then raise the alarm here
					if (prefixes.size()>1){
					asAlarm = key;
					toReturn.put(key, loc.getLocalisedMessageText("news.misc.anomalous")+
							" "+loc.getLocalisedMessageText("news.regions.prefix")+" "+prefixes.toString());
					}
					
				} else {
			
				double maxRatio = 0;
				
				// figure out if they overlap
				
				// stores only peers that actually saw a problem at this time
				HashSet<Integer> peersSeeingProblem = new HashSet<Integer>();
				
				// stores all peers online an their detection probabilities
				HashMap<Integer, Double> perPeerProbs = new HashMap<Integer, Double>();
				
				perPeerProbs.put(AnomalyRecord.getMyId(), ddps.detectionProb);
				for (AnomalyRecord ar : anomalies.values()){
					if (ar.detectionProb<=0 || ar.detectionProb>1) continue;
					peersSeeingProblem.add(ar.id);
				}
				if (peersSeeingProblem.size()<3) return toReturn; // TODO make this a variable
				
				// add all the independent probs
				perPeerProbs.putAll(ddps.globalIndepProbs.get(key));
				
				HashMap<Integer, BigDecimal> unionProbs = 
					Util.findUnionProbability(perPeerProbs.size(), perPeerProbs, 25);
				
				double oneProb = unionProbs.get(0).doubleValue();
				if (DEBUG) System.out.println("Odds of one peer seeing it: "+unionProbs.get(0).doubleValue());
				for (Entry<Integer, BigDecimal> ent2 : unionProbs.entrySet()){
					if (ent2.getKey()==0) continue;
					if (ent2.getKey()>peersSeeingProblem.size()) continue; // not actually seeing this many peers
					
					// update observed rate of corroboration
					if (ddps.corroboratedCounts.get(ent2.getKey())==null){
						ddps.corroboratedCounts.put(ent2.getKey(), new HashSet<Long>());						
					}
					ddps.corroboratedCounts.get(ent2.getKey()).add(currentBucket);
					double corrobProb = ddps.corroboratedCounts.get(ent2.getKey()).size()/
										(((double)uptime)/(bucketSize/1000));
					
					// compare to independent rate of observation
					if (ent2.getValue().doubleValue()>0){
						if (DEBUG) System.out.println("Ratio for "+(ent2.getKey()+1)
								+": "+(oneProb/ent2.getValue().doubleValue()));
						double likelihoodRatio = corrobProb/ent2.getValue().doubleValue();
						if (DEBUG) System.out.println("Likelihood Ratio for "+(ent2.getKey()+1)
								+": "+likelihoodRatio);	
						if (likelihoodRatio > maxRatio) maxRatio = likelihoodRatio;
						
					}
				}
				
				if (maxRatio > alarmThreshold){
					prefixAlarm = key;
					toReturn.put(key, 
							((int)maxRatio) + " "+ loc.getLocalisedMessageText("news.likelihood"));
					
					// prepare anomlay record for notifying AS-level issues
					AnomalyRecord  ar = new AnomalyRecord();
					ar.id = Util.convertIpToInt(key.replace("Prefix:", "").split("/")[0]);
					ar.detectionProb = (float)maxRatio;
					
					// insert AS-level anomaly
					// TODO if this is really what you want to do, key!=ASN at this point...
					ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					try {
						ar.serialize(baos);					
						baoss.add(baos);
						dht.write(key, baoss, 
								NEWSConfiguration.getInstance().getSampleInterval(), awa);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}

				} // end case prefix anomaly
			if (prefixAlarm.length()!=0){
				if (prefixAlarmRaised){
					// suppress if alarm was already raised
					toReturn.remove(prefixAlarm);
					prefixAlarm = "";
					Main.getLogger().log("Suppressing duplicate alarm on prefix...");
				}
				prefixAlarmRaised = true;
			}
			// don't show alarm for AS if prefix isn't affected...
			if (asAlarm.length()>0 ) {
//				if (!prefixAlarmRaised){
//					toReturn.remove(asAlarm);
//					asAlarm = "";
//				} else{
					if (asAlarmRaised){
						// suppress if alarm was already raised
						toReturn.remove(asAlarm);
						asAlarm = "";
						Main.getLogger().log("Suppressing duplicate alarm on AS...");
					}
					asAlarmRaised = true;
//				}
			}
			// don't show alarm for country if AS isn't affected
			if (countryAlarm.length()>0 ){
				if (!asAlarmRaised){
					toReturn.remove(countryAlarm);
					countryAlarm = "";
				} else{
					if (countryAlarmRaised){
						// suppress if alarm was already raised
						toReturn.remove(countryAlarm);
						countryAlarm = "";
						Main.getLogger().log("Suppressing duplicate alarm on country...");
					}
					countryAlarmRaised = true;
				}
			}
			
			return toReturn;
		}

//		@Override
		public void handleTimeout(IDistributedDatabaseEvent event) {
			Main.getLogger().log("Read timed out!");
			
		}
		
	}
	
	private class DetectionDataPerSetting{
		/** counts of actual problems per region (prefix, AS, country), grouped first 
		 * by region, then by time bucket
		 */
		private HashMap<String, HashMap<Long, HashMap<Integer, AnomalyRecord>>> globalAnomalyCounts;
		/** counts of probabilities per region (prefix, AS, country), grouped first 
		 * by region, then map of per-session id to record
		 */
		protected HashMap<String, HashMap<Integer, Double>> globalIndepProbs;
		/** map of region to set of times where already anomalous */
		private HashMap<String, HashSet<Long>> foundAnomalies = new HashMap<String, HashSet<Long>>();
	    /** number of sequential anomalies detected */
		private int recentAnomalies = 0;
		int anomalyEventCount = 0;
		/** map of number of corroborating peers to set of buckets having corroboration*/
		HashMap<Integer, HashSet<Long>> corroboratedCounts;
		
		double detectionProb;

		public DetectionDataPerSetting(){
			corroboratedCounts = new HashMap<Integer, HashSet<Long>>();
			foundAnomalies = new HashMap<String, HashSet<Long>>();
			globalAnomalyCounts = new HashMap<String, HashMap<Long, HashMap<Integer, AnomalyRecord>>>();
			globalIndepProbs = new HashMap<String, HashMap<Integer, Double>>();
		}
	}

	private static final boolean DEBUG = false;
	protected static final boolean TESTING = false;
	private static final boolean INSERT_TEST_ANOMALIES = false;

	/** minimum time between alerts from NEWS */
	private static final long ALERT_SQUELCH_TIME = 5*60*1000;
	/** maximum history for anomalies, currently 1 day */
	private static final long MAX_HISTORY_TIME = 24*60*60*1000;

	/** map of names to moving average that are being watched */
	HashMap<String, ArrayList<MovingAverage>> movingAverages;
	
	/** map of properties to map of metric to MA*/
	HashMap<MovingAverageHistory, HashMap<String, MovingAverage>> perSettingAverages;
	
	/** map of properties to detection data */
	HashMap<MovingAverageHistory, DetectionDataPerSetting> perSettingDetectionData;
	
	/** map of names to moving average that are being watched */
	HashMap<String, Sampler> samplers;
	
	/** set of names of metrics to use */
	HashSet<String> names;
	
	/** timer for controlling frequency */
	ITimer myTimer;

	private int anomCountThreshold = 2;

	private IDistributedDatabase dht;
	
	private AnomalyWriteAction awa;
	private AnomalyReadAction ara;
	
	/** window of time to consider the same "time" for anomalies (milliseconds)*/
	protected int bucketSize = 5*60*1000;
	
	/** current time region for accumulating alerts */
	private long currentBucket = -1;
	
	/** minimum number of alerts to see before raising alarm */
	public int alarmThreshold = 5;
	/** sample frequency (seconds) */
	private int sampleFrequency;
	/** last time alert was shown */
	private long lastAlertTime = -1;

	private static AnomalyDetector self;
	
	private ArrayList<AnomalyListener> listeners;

	private HashMap<String, Integer> metricIds;
//	protected double lastStdDev=-1;
	private boolean prefixAlarmRaised;
	private boolean asAlarmRaised;
	private boolean countryAlarmRaised;
	private boolean initComplete = false;
	
	long startTime;
	long lastTime;
	int uptime;
	int numObservations = 0;
	
	/** number of active downloads */
	int numDownloads = 0;
	/** number of active uploads */
	int numUploads = 0;
	private PluginInterface pi;
	
	public AnomalyDetector(ITimer anomalyTimer){	
		
		self = this;
		myTimer = anomalyTimer;
		sampleFrequency = NEWSConfiguration.getInstance().getSampleInterval();
		startTime = System.currentTimeMillis();
		lastTime = System.currentTimeMillis();
		
		
		names = new HashSet<String>();
		names.add("totalDlRate");
		names.add("totalUlRate");
		names.add("perConDlRate");
		names.add("perConUlRate");
		names.add("availability-dl");
		names.add("availability-seed");
		names.add("numPeers-dl");
		names.add("numPeers-seed");
		names.add("numLeechers");
		names.add("numSeeds");
		names.add("rtt");
		names.add("loss");
		names.add("rst");
		
		metricIds = new HashMap<String, Integer>();
		metricIds.put("totalDlRate", 1);
		metricIds.put("totalUlRate", 2);
		metricIds.put("perConDlRate", 4);
		metricIds.put("perConUlRate", 8);
		metricIds.put("availability-dl", 16);
		metricIds.put("availability-seed", 32);
		metricIds.put("numPeers-dl", 64);
		metricIds.put("numPeers-seed", 128);
		metricIds.put("numLeechers", 256);
		metricIds.put("numSeeds", 512);
		metricIds.put("rtt", 1024);
		metricIds.put("loss", 2048);
		metricIds.put("rst", 4096);
		
		
		// fill this in:
		movingAverages = new HashMap<String, ArrayList<MovingAverage>>();
		
		// first create all the moving averages
		// this is one per window and set of stdDevs
		for (String name : names){
			for (int windowSize : AnomalyProperties.getInstance().getWindowSizes()){
				MovingAverageConfig mac = AnomalyProperties.getInstance().getConfig(name);
				if (mac!=null) {
					mac.windowSize = windowSize;
					MovingAverage ma = new MovingAverage(mac);
					if (movingAverages.get(name)==null) movingAverages.put(name, new ArrayList<MovingAverage>());
					movingAverages.get(name).add(ma);
				}
			}
		}
		
		// then associate moving averages with histories
		if (NetworkAdminASNLookupImpl.getInstance().getASNInfo()!=null) buildHistories();

		samplers = new HashMap<String, Sampler>();
		setupSamping();
		
		awa = new AnomalyWriteAction();
		ara = new AnomalyReadAction();		
		updateCurrentBucket();
		alarmThreshold = NEWSConfiguration.getInstance().getAlarmThreshold();
		anomCountThreshold = NEWSConfiguration.getInstance().getAnomCountThreshold();
		listeners = new ArrayList<AnomalyListener>();
		initComplete = true;
		pi = MainGeneric.getPluginInterface();
	}


	/**
	 * 
	 */
	private void buildHistories() {
		perSettingAverages = new HashMap<MovingAverageHistory, HashMap<String, MovingAverage>>();		
		perSettingDetectionData = new HashMap<MovingAverageHistory, DetectionDataPerSetting>();
		for (MovingAverageHistory mah : AnomalyProperties.getInstance().getMovingAverageHistories()){
			perSettingAverages.put(mah, new HashMap<String, MovingAverage>());
			for (String name : names){
				for (MovingAverage ma : movingAverages.get(name)){
					if (ma.getWindowSize()==mah.windowSize){
						perSettingAverages.get(mah).put(name, ma);
						break;
					}
				}
			}
			perSettingDetectionData.put(mah, new DetectionDataPerSetting());
		}
	}


	private boolean updateCurrentBucket() {
		long localTime = Util.currentNEWSTime();
		long myCurrentBucket = localTime - localTime%bucketSize;
		if (myCurrentBucket!=currentBucket) {			
			currentBucket = myCurrentBucket;
			// reset alarm status
			prefixAlarmRaised = false;
			asAlarmRaised = false;
			countryAlarmRaised = false;
			return true;
		} else return false;
	}


	private void setupSamping() {
		final PluginInterface pi = MainGeneric.getPluginInterface();
		for (String name : names){
			if (movingAverages.get(name)==null) continue;
			ArrayList<MovingAverage> mas = movingAverages.get(name);
			if (name.equals("totalDlRate")){
				// TODO use difference in bytes divided by difference in time
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						boolean isDownloading = false;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_DOWNLOADING) isDownloading = true;
						}
						if (!isDownloading) for (MovingAverage ma : myAverage) ma.removeSample();
						else for (MovingAverage ma : myAverage) ma.addSample(pi.getDownloadManager().getStats().getDataReceiveRate());
					}});
			} else if (name.equals("totalUlRate")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						for (MovingAverage ma : myAverage) ma.addSample(pi.getDownloadManager().getStats().getDataSendRate());
					}});
			} else if (name.equals("perConDlRate")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumRates =0;
						int count = 0;
						// only if downloading
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()!=Download.ST_DOWNLOADING) continue;
							if (d.getPeerManager()!=null){
								for (Peer p : d.getPeerManager().getPeers()){
//									if (p.isInteresting()){
										if (p.getStats().getDownloadAverage()>0){
											sumRates+=p.getStats().getDownloadAverage();
											count++;
										}
//									}
								}
							}
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumRates/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("perConUlRate")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumRates =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getPeerManager()!=null){
								for (Peer p : d.getPeerManager().getPeers()){
//									if (p.isInterested()){
										if (p.getStats().getUploadAverage()>0){
											sumRates+=p.getStats().getUploadAverage();
											count++;
										}
//									}
								}
							}
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumRates/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("availability-dl")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumAvail =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_DOWNLOADING){
								sumAvail+=d.getStats().getAvailability();
								count++;
							}
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumAvail/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("availability-seed")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumAvail =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_SEEDING){
								sumAvail+=d.getStats().getAvailability();
								count++;
							}
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumAvail/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("numPeers-dl")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_DOWNLOADING){
								if (d.getPeerManager()==null) continue;
								sumPeer+=d.getPeerManager().getPeers().length;
								count++;
							}
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumPeer/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("numPeers-seed")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_SEEDING){
								if (d.getPeerManager()==null) continue;
								sumPeer+=d.getPeerManager().getPeers().length;
								count++;
							}
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumPeer/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("numLeechers")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							
							if (d.getPeerManager()==null) continue;
							sumPeer+=d.getPeerManager().getStats().getConnectedLeechers();
							count++;
							
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumPeer/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("numSeeds")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){

								if (d.getPeerManager()==null) continue;
								sumPeer+=d.getPeerManager().getStats().getConnectedSeeds();
								count++;
							
						}												
						if (count>0) for (MovingAverage ma : myAverage) ma.addSample(sumPeer/count);
						else for (MovingAverage ma : myAverage) ma.removeSample();
					}});
			} else if (name.equals("rtt")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						for (MovingAverage ma : myAverage) ma.addSample(PingManager.getInstance().getRttSample());
					}});
			}  else if (name.equals("loss")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						for (MovingAverage ma : myAverage) ma.addSample(TraceRouteRunner.getInstance().getLossSample());
					}});
			}  else if (name.equals("rst")){
				samplers.put(name, new Sampler(mas, name, pi){

					@Override
					public void sample() {
						try {
							for (MovingAverage ma : myAverage) ma.addSample(RSTPacketStats.getInstance().getMostRecent().deltaConnReset);
//						System.out.println("RST, found: "+RSTPacketStats.getInstance().getMostRecent().deltaConnReset);
						} catch (Exception e){
							// do nothing
						}
					}});
			}
			}
		
		
	}


//	@Override
	public void perform(ITimerEvent event) {
		if (!initComplete) return;
		if (perSettingAverages==null && 
				NetworkAdminASNLookupImpl.getInstance().getASNInfo()!=null) buildHistories();
		else if (NetworkAdminASNLookupImpl.getInstance().getASNInfo()==null) return;
	
		NetworkAdminASNImpl asnInfo = NetworkAdminASNLookupImpl.getInstance().getASNInfo();
		String keys[] = new String[]{"Prefix:"+asnInfo.getBGPPrefix(), "AS:"+asnInfo.getAS(), "Country:"+asnInfo.getCountry()};

		// make sure that there is a moving average history for each network
		for (String key : keys){
			boolean found = false;
			for (MovingAverageHistory mah : perSettingAverages.keySet()){
				if (mah.networkName.equals(key)) {
					found = true;
					break;
				}
			}
			if (!found) createMovingAverageHistory(key);
		}
		
		numObservations ++;
		updateUptime();
		LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();

		// check for updates
		// TODO change this to check for parameter ranges, restore commented code
		// TODO check for changes to the set of active settings based on 
		// the rate at which anomalies are detected
		// TODO periodically check if network changes
		for (String name: names){
//			if (movingAverages.get(name)!=null && !movingAverages.get(name).configEquals(					
//					AnomalyProperties.getInstance().getConfig(name))){
//				System.out.println("New configuration for "+name+"!");
//				
//				movingAverages.put(name, new MovingAverage(
//						AnomalyProperties.getInstance().getConfig(name)));
//				samplers.get(name).setMovingAverage(movingAverages.get(name));
//			}
//			if (lastStdDev<0) {
//				lastStdDev = AnomalyProperties.getInstance().getConfig(name).stdDevThresh;
//			}
		}
		alarmThreshold = NEWSConfiguration.getInstance().getAlarmThreshold();
		anomCountThreshold = NEWSConfiguration.getInstance().getAnomCountThreshold();
		
		for (String name : names){
//			if (DEBUG) System.out.println("Sampling "+name);
			if (samplers.get(name)!=null) samplers.get(name).sample();
		}
		
		// iterate over each detection setting
		int emptyEntryCount = Integer.MAX_VALUE;
		boolean shouldAnnounce = false;
		for (Entry<MovingAverageHistory, HashMap<String, MovingAverage>> ent : 
			perSettingAverages.entrySet()){
			
			// check that the history corresponds to one of our current networks
			boolean skip = true;
			for (String key : keys){
				if (ent.getKey().networkName.equals(key)) skip = false;
			}
			if (skip) continue;
			
			HashMap<String, MovingAverage> myMovingAverages = ent.getValue();

			// TODO check to see if the history is in the current set of locations
			
			int anomalyCount = 0;
			int rateCount = 0;
			int window=0;
			ArrayList<String> anomalousNames = new ArrayList<String>();
			int anomalyTypes = 0;

		
		// these two values allow NEWS to publish the minimal std dev seen in 
		// the signal. This could be higher than the threshold, allowing other 
		// peers with higher thresholds to use the anomaly information.
		double minStdDev = Double.MAX_VALUE;
		double currentStdDev;
		
		
		for (String name : names){
			
			// check "isAnomalous" with the properties settings currently being inspected
			if (myMovingAverages.get(name)!=null && myMovingAverages.get(name).isAnomalous(ent.getKey().stdDevThresh)){				
				if (name.contains("Rate")){
					// make sure a significant drop has occurred
					if (myMovingAverages.get(name).getDeviationPercent()<
							NEWSConfiguration.getInstance().getPctDiffTrigger()/100.0) continue;
					
					// check for change in active upload/downloads
					int currentDownloads = 0;
					for (Download d : pi.getDownloadManager().getDownloads()){
						if (d.getState()==Download.ST_DOWNLOADING) currentDownloads++;
					}
					if (currentDownloads < numDownloads) {
						numDownloads = currentDownloads;
						continue;
					}
					numDownloads = currentDownloads;
					
					int currentUploads = 0;
					for (Download d : pi.getDownloadManager().getDownloads()){
						if (d.getState()==Download.ST_SEEDING) currentUploads++;
					}
					if (currentUploads < numUploads) {
						numUploads = currentUploads;
						continue;
					}
					numUploads = currentUploads;
					
					
					window = myMovingAverages.get(name).getWindowSize();
//					lastStdDev = myMovingAverages.get(name).getStdDevThresh();
					rateCount++;
					currentStdDev = myMovingAverages.get(name).getAnomalousStdDev();
					if (currentStdDev < minStdDev) minStdDev = currentStdDev;
				} else { // not a rate
					currentStdDev = myMovingAverages.get(name).getAnomalousStdDev();
					if (currentStdDev < minStdDev) minStdDev = currentStdDev;
				}
				anomalyCount++;
				anomalousNames.add(loc.getLocalisedMessageText("news.sensors."+name));
				anomalyTypes |= metricIds.get(name);
			} else if (myMovingAverages.get(name)!=null && !myMovingAverages.get(name).isValid()){
				if (myMovingAverages.get(name).entryCount()!=0){
					if (emptyEntryCount>myMovingAverages.get(name).getEmptyEntryCount())
						emptyEntryCount = myMovingAverages.get(name).getEmptyEntryCount();
					shouldAnnounce=true;
				}
			} else if (myMovingAverages.get(name)!=null && 
					myMovingAverages.get(name).isValid()) shouldAnnounce=false;
		}
		
		
		DetectionDataPerSetting ddps = perSettingDetectionData.get(ent.getKey());
		if (anomalyCount>=anomCountThreshold  && rateCount>0){
			ddps.anomalyEventCount++;
			ent.getKey().addAnomalyHistory(true);
			String intro = loc.getLocalisedMessageText("news.messages.alert");
			if (canDisplayAlert()) Main.getLogger().logAlert(LoggerChannel.LT_INFORMATION, 
					"["
					+ DateFormat.getDateTimeInstance().format(
							new Date()) + "] "+
							intro+"\n\n"
					+ loc.getLocalisedMessageText("news.misc.anomalous") + 
					" " + loc.getLocalisedMessageText("news.misc.sensors")+ ": "+
					anomalousNames);
			else if (DEBUG) System.out.println("Suppressing log...");
			ddps.recentAnomalies++;
			if (dht!=null || MainGeneric.getDistributedDatabase()!=null){
				if (dht == null) dht = MainGeneric.getDistributedDatabase();
				
				AnomalyRecord ar = new AnomalyRecord();
				ar.windowSize = window;
				ar.stdDevThresh = minStdDev;
				ar.numAnomalies = anomalyCount;
				ar.id = AnomalyRecord.getMyId();
				ar.anomalyTypes = anomalyTypes;
				ar.uptime = ent.getKey().getHistoryDuration();
				ar.detectionProb = (float)ent.getKey().getDetectionProb(ddps.recentAnomalies, numObservations);
				ddps.detectionProb = ar.detectionProb;
				Main.getLogger().log("Writing anomaly for id "+ar.id);
				Statistics.getInstance().addAnomalyEvent(ar);
				// notify listeners
				for (AnomalyListener al : listeners) al.alertRaised(ar);
				
				
				if (updateCurrentBucket()){
					for (String key : keys){
						if (ddps.globalAnomalyCounts.get(key)==null) 
							ddps.globalAnomalyCounts.put(key, new HashMap<Long, HashMap<Integer, AnomalyRecord>>());
						if (ddps.globalAnomalyCounts.get(key).get(currentBucket)==null)  
							ddps.globalAnomalyCounts.get(key).put(currentBucket, new HashMap<Integer, AnomalyRecord>());
						ddps.globalAnomalyCounts.get(key).get(currentBucket).put(AnomalyRecord.getMyId(), ar);
						ArrayList<Long> toRemove = new ArrayList<Long>();
						for (Long t : ddps.globalAnomalyCounts.get(key).keySet()){
							if (currentBucket-t>MAX_HISTORY_TIME) toRemove.add(t);
						}
						for (long t : toRemove) ddps.globalAnomalyCounts.get(key).remove(t);
					}
				}
				
				ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();
				for (String key : keys){
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					baoss.clear();
					try {
						ar.serialize(baos);					
						baoss.add(baos);
						dht.write(key, baoss, 
								NEWSConfiguration.getInstance().getSampleInterval(), awa);
					} catch (IOException e) {						
						e.printStackTrace();
					}
				}
				
				final AnomalyRecord ar2 = ar;
				for (String key : keys){
					dht.read(key, "Checking for anomalies in region: "+key + " @"+Util.currentNEWSTime(), 
							NEWSConfiguration.getInstance().getSampleInterval(), 
							ara, dht.getOptions().getExhaustiveRead());
					final String key2 = key;
					// schedule delayed re-reads and re-writes
					for (long offset = 10*60*1000; offset <= 60*60*1000; offset+=10*60*1000){
						// re-read
						myTimer.addEvent(System.currentTimeMillis()+offset, new ITimerEventPerformer(){
							// TODO create new object, eliminate duplicate reads per time window
							public void perform(ITimerEvent event) {
								dht.read(key2, "Checking for anomalies in region: "+key2 + " @"+Util.currentNEWSTime(), 
										NEWSConfiguration.getInstance().getSampleInterval(), 
										ara, dht.getOptions().getExhaustiveRead());
								
							}});
						// re-write
						myTimer.addEvent(System.currentTimeMillis()+offset, new ITimerEventPerformer(){
							// TODO create new object, eliminate duplicate reads per time window
							public void perform(ITimerEvent event) {
								ByteArrayOutputStream baos = new ByteArrayOutputStream();
								ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();
								try {
									baoss.clear();
									ar2.serialize(baos);					
									baoss.add(baos);
									// TODO schedule re-writes similar to re-reads below
									dht.write(key2, baoss, 
											NEWSConfiguration.getInstance().getSampleInterval(), awa);
								} catch (IOException e) {						
									e.printStackTrace();
								}	
							}});
					}
					
				}
				
			}
		} else {
			ddps.recentAnomalies =0;
			ent.getKey().addAnomalyHistory(false);
		}
		} // end iterate over each anomaly detection setting
		
		// print out number to go
		if (shouldAnnounce && emptyEntryCount<Integer.MAX_VALUE){
			String finalText = emptyEntryCount + " more " +
			"samples to go...";
			if (emptyEntryCount==1){
				finalText = "will be ready to go after next sample!";
			}
			Main.getLogger().log("["
					+ DateFormat.getDateTimeInstance().format(
							new Date()) + "] "+
					"NEWS is still warming up, "+ 
					finalText);
		}
		
		// check sample rate
		if (!TESTING && sampleFrequency!=NEWSConfiguration.getInstance().getSampleInterval()){			
			if (DEBUG) System.out.println("Setting sample frequency to "+
					NEWSConfiguration.getInstance().getSampleInterval() + 
					" seconds!");
			event.cancel();
			sampleFrequency = NEWSConfiguration.getInstance().getSampleInterval();
			myTimer.addPeriodicEvent(sampleFrequency*1000, this);
		} else NEWSConfiguration.getInstance().setSampleInterval(sampleFrequency);
		
		if (INSERT_TEST_ANOMALIES) doTestAnomalies();
		
	} // end perform
	
	private void createMovingAverageHistory(String networkName) {
		ArrayList<MovingAverageHistory> mahs = AnomalyProperties.getInstance().createMovingAverageHistory(networkName);
		for (MovingAverageHistory mah : mahs){
			perSettingAverages.put(mah, new HashMap<String, MovingAverage>());
			for (String name : names){
				for (MovingAverage ma : movingAverages.get(name)){
					if (ma.getWindowSize()==mah.windowSize){
						perSettingAverages.get(mah).put(name, ma);
						break;
					}
				}
			}
			perSettingDetectionData.put(mah, new DetectionDataPerSetting());
			perSettingDetectionData.get(mah).detectionProb = mah.getDetectionProb(0, numObservations);
		}
	}


	private void updateUptime() {
		long diff = System.currentTimeMillis()-lastTime;
		// just add one window if time jumps
		if (diff < 0 || diff > 2*1000*NEWSConfiguration.getInstance().getSampleInterval()){
			uptime+= NEWSConfiguration.getInstance().getSampleInterval();
		} else uptime += (int)(diff/1000);
		lastTime = System.currentTimeMillis();
	
	}


	private void doTestAnomalies() {
		int NUM_REPS = 5;
		
		DetectionDataPerSetting ddps = new DetectionDataPerSetting();
		
		for (int i = 0; i < NUM_REPS; i++){
		if (dht!=null || MainGeneric.getDistributedDatabase()!=null){
			if (dht == null) dht = MainGeneric.getDistributedDatabase();
			NetworkAdminASNImpl asnInfo = NetworkAdminASNLookupImpl.getInstance().getASNInfo();
		String keys[] = new String[]{"Prefix:"+asnInfo.getBGPPrefix(), 
				//"AS:"+asnInfo.getAS(), 
				/*"Country:"+asnInfo.getCountry()*/};
		
		AnomalyRecord ar = new AnomalyRecord();
		ar.windowSize = 25;
		ar.stdDevThresh = 2.5;
		ar.numAnomalies = 3;
		ar.id = (int)(Math.random()*Integer.MAX_VALUE);
		ar.anomalyTypes = Integer.MAX_VALUE;
		ar.uptime = uptime;
		ar.detectionProb = (float)(0.5*Math.random());//(0.001+Math.random()*0.05);		
		Main.getLogger().log("Writing anomaly for id "+ar.id);
//		Statistics.getInstance().addAnomalyEvent(ar);
		// notify listeners
//		for (AnomalyListener al : listeners) al.alertRaised(ar);
		
		if (updateCurrentBucket()){
			for (String key : keys){
				if (ddps.globalAnomalyCounts.get(key)==null) 
					ddps.globalAnomalyCounts.put(key, new HashMap<Long, HashMap<Integer, AnomalyRecord>>());
				if (ddps.globalAnomalyCounts.get(key).get(currentBucket)==null)  
					ddps.globalAnomalyCounts.get(key).put(currentBucket, new HashMap<Integer, AnomalyRecord>());
				ddps.globalAnomalyCounts.get(key).get(currentBucket).put(AnomalyRecord.getMyId(), ar);
				ArrayList<Long> toRemove = new ArrayList<Long>();
				for (Long t : ddps.globalAnomalyCounts.get(key).keySet()){
					if (currentBucket-t>MAX_HISTORY_TIME) toRemove.add(t);
				}
				for (long t : toRemove) ddps.globalAnomalyCounts.get(key).remove(t);
			}
		}
		
		ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();
		for (String key : keys){
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			try {
				if (key.contains("AS")) ar.detectionProb+= 20;
				ar.serialize(baos);					
				baoss.add(baos);
				dht.write(key, baoss, 
						NEWSConfiguration.getInstance().getSampleInterval(), awa);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		// TODO make this more configurable
		//if (numObservations<20) continue;
		
		for (String key : keys){
			dht.read(key, "Checking for anomalies in region: "+key + " @"+Util.currentNEWSTime(), 
					NEWSConfiguration.getInstance().getSampleInterval(), 
					ara, dht.getOptions().getExhaustiveRead());
		}
		}
		} // end loop of repetitions
		
	}


	private boolean canDisplayAlert(){
		if (!NEWSConfiguration.getInstance().isSuppressWarnings() && 
				Util.currentNEWSTime()-lastAlertTime > ALERT_SQUELCH_TIME){
			lastAlertTime = Util.currentNEWSTime();
			return true;
		}
		return false;
		
	}
	
	public static AnomalyDetector getInstance(){
		return self;
	}
	
	public void addListener(AnomalyListener al){
		listeners.add(al);
	}
	
	public void removeListener(AnomalyListener al){
		listeners.remove(al);
	}


	public void writeNoop() {
		if (dht!=null || MainGeneric.getDistributedDatabase()!=null){
			if (dht == null) dht = MainGeneric.getDistributedDatabase();
			NetworkAdminASNImpl asnInfo = NetworkAdminASNLookupImpl.getInstance().getASNInfo();
			String key = "Prefix:"+asnInfo.getBGPPrefix();

			// loop over each detection setting
			for (Entry<MovingAverageHistory, DetectionDataPerSetting> ent : perSettingDetectionData.entrySet()){
				// create corresponding record
				AnomalyRecord ar = new AnomalyRecord();
				ar.windowSize = ent.getKey().windowSize;
				ar.stdDevThresh = ent.getKey().stdDevThresh;
				ar.numAnomalies = 0;
				ar.id = AnomalyRecord.getMyId();
				ar.anomalyTypes = 0; // makes this a "NOOP" record
				ar.uptime = uptime;
				ar.detectionProb = (float)ent.getKey().getDetectionProb(0, numObservations);//(0.001+Math.random()*0.05);	

				// put in DHT
				ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();		
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				try {				
					ar.serialize(baos);					
					baoss.add(baos);
					dht.write(key, baoss, 
							NEWSConfiguration.getInstance().getSampleInterval(), awa);
				} catch (IOException e) {
					e.printStackTrace();
				}

			}
		}
		
	}

}
