/**
 * NEWS Project
 *
 * File:         AnomalyProperties.java
 * RCS:          $Id: AnomalyProperties.java,v 1.8 2009/11/29 19:21:54 drc915 Exp $
 * Description:  AnomalyProperties class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 7, 2008 at 10:24:10 AM
 * Language:     Java
 * Package:      edu.northwestern.news.anomaly
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.news.anomaly;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Properties;

import edu.northwestern.news.NEWSConfiguration;
import edu.northwestern.news.asn.NetworkAdminASNImpl;
import edu.northwestern.news.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.news.util.Util;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The AnomalyProperties class ...
 */
public class AnomalyProperties {

	public static class MovingAverageConfig{
		int windowSize;
		Double stdDevThresh[];
		int anomalyDirection;
		
		public String toString(){
			return windowSize + ";"+Util.join("|", stdDevThresh)+";"+anomalyDirection;
		}
	}


	public static final boolean DEBUG = false;
	public static final String ANOMALY_DETECTION_PROPERTIES = "news-detection-config.properties";
	private static final String ANOMALY_HISTORY_PROPERTIES = "anomaly-history.properties";
	
	/** supports multiple concurrent MAs and persistent storage of detection rates */
	public static class MovingAverageHistory{		
		int windowSize;
		double stdDevThresh;
		String networkName;
		BitSet history; 
		long lastRecorded; 
//		long endTime;
		String hashString;
		int currentSessionAnoms = 0;
		int historyAnoms = -1;
		int historySize = 0;
		ArrayList<Boolean> currentSession = new ArrayList<Boolean>();
		static int MAX_HISTORY = (30*24*60*60)/NEWSConfiguration.getInstance().getSampleInterval();
		
		public MovingAverageHistory(){
			
		}
		
		public int hashCode() {
			if (hashString==null){
				hashString = windowSize+stdDevThresh+networkName;
			}			
			return hashString.hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if (obj instanceof MovingAverageHistory){
				MovingAverageHistory other = (MovingAverageHistory) obj;
				return other.windowSize == this.windowSize 
					&& other.stdDevThresh==this.stdDevThresh && 
					other.networkName.equals(networkName);
			}
			return false;
		}

		public float getDetectionProb(int consecutiveEvents, int numObservations) {
			if (historyAnoms<0) historyAnoms = history.cardinality();
			if (consecutiveEvents>1) currentSessionAnoms--; // consecutive anoms treated as one
			double totalSlots = history.size()+numObservations;
			double totalAnoms = currentSessionAnoms+historyAnoms;
			return (float)(totalAnoms/totalSlots);
		}
		
		public void addAnomalyHistory(boolean isAnomaly){
			if (currentSession.size()>MAX_HISTORY) currentSession.remove(0);
			currentSession.add(isAnomaly);
			if (isAnomaly)currentSessionAnoms++;
			
		}
		
		
		// TODO load from file
		public void outputToString(Properties props, int id){
			
			// first merge histories
			int totalHistory = historySize+currentSession.size();
			int startOffset = 0;
			if (totalHistory>MAX_HISTORY) {
				startOffset = totalHistory-MAX_HISTORY;
				totalHistory = MAX_HISTORY;
			}
			
			BitSet temp = new BitSet(totalHistory);
			int tempIndex = 0;
			for (int i = startOffset; i < historySize; i++ ){
				if (history.get(i)) temp.set(tempIndex);
				tempIndex++;
			}
			for (boolean b : currentSession){
				if (b) temp.set(tempIndex);
				tempIndex++;
			}
			historySize = totalHistory;
			history = temp;
			
			
			props.put(id+".windowSize",windowSize+"");
			props.put(id+".stdDevThresh",stdDevThresh+"");
			props.put(id+".networkName",networkName+"");
			props.put(id+".recordedAt",Util.currentGMTTime()+"");
			props.put(id+".history", buildHistoryString());

		}

		private String buildHistoryString() {
			StringBuilder sb = new StringBuilder();
			sb.append(historySize+":");
			int nextSetBit = history.nextSetBit(0);
			while (nextSetBit>0){
				sb.append(nextSetBit);
				nextSetBit = history.nextSetBit(nextSetBit+1);
				if (nextSetBit>0) sb.append(",");
			}
			return sb.toString();
		}
		
		public static HashMap<String, ArrayList<MovingAverageHistory>> loadFromProperties(Properties props){
			// load them all for now, use only those that we care about
			HashMap<Integer, MovingAverageHistory> tempHolder = new HashMap<Integer, MovingAverageHistory>();
			for (Object o : props.keySet()){
				try {
				String parts[] = ((String)o).split("\\.");
				int id = Integer.parseInt(parts[0]);
				MovingAverageHistory mah = tempHolder.get(id);
				if (mah==null){
					tempHolder.put(id, new MovingAverageHistory());
					mah = tempHolder.get(id);
				}
				if (parts[1].equals("windowSize")){
					mah.windowSize=Integer.parseInt(props.getProperty((String)o));
				} else if (parts[1].equals("stdDevThresh")){
					mah.stdDevThresh=Double.parseDouble(props.getProperty((String)o));
				}else if (parts[1].equals("networkName")){
					mah.networkName = props.getProperty((String)o);
				}else if (parts[1].equals("recordedAt")){
					mah.lastRecorded=Long.parseLong(props.getProperty((String)o));
				}else if (parts[1].equals("history")){
					String hist[]= props.getProperty((String)o).split("\\:");
					mah.history = new BitSet(Integer.parseInt(hist[0]));
					mah.historySize = Integer.parseInt(hist[0]);
					if (hist.length>1 && hist[1].trim().length()>0) {
						for (String set : hist[1].split(",")){					
							mah.history.set(Integer.parseInt(set));
						}
					}
				}		
				} catch (NumberFormatException e){
					if (DEBUG) e.printStackTrace();
				}
			}
			
			// check for freshness and add to hashmap
			HashMap<String, ArrayList<MovingAverageHistory>> toReturn = new 
			HashMap<String, ArrayList<MovingAverageHistory>>();
			for (MovingAverageHistory mah : tempHolder.values()){
				if (Math.abs(Util.currentGMTTime()-mah.lastRecorded)<30L*24*60*60*1000){
					if (toReturn.get(mah.networkName)==null) 
						toReturn.put(mah.networkName, new ArrayList<MovingAverageHistory>());
					toReturn.get(mah.networkName).add(mah);
				}
			}
			
			
			return toReturn;
		}

		public int getHistoryDuration() {			
			return Math.min(MAX_HISTORY, history.size()+currentSession.size())*
			NEWSConfiguration.getInstance().getSampleInterval();
		}
		
	}


	private static AnomalyProperties self;
	private static double stdDevIncr = 0.1;
	
	private HashMap<String, MovingAverageConfig> configs;
	
	/** map of network name to histories */
	private HashMap<String, ArrayList<MovingAverageHistory>> histories;
	
	private HashSet<Double> stdDevs = new HashSet<Double>();
	private HashSet<Integer> windowSizes = new HashSet<Integer>();
	
	String namesA[] = null;
	String directionsA[] = null;
	
	public MovingAverageConfig getConfig(String name){
		return configs.get(name);
	}

	public AnomalyProperties(){
		configs = new HashMap<String, MovingAverageConfig>();
		histories = new HashMap<String, ArrayList<MovingAverageHistory>>();
		
	}
	
	public static AnomalyProperties getInstance() {
		if (self == null){
			self = new AnomalyProperties();
		} 
		return self;
	}

	public void loadFromProperties(Properties anom) {
		ArrayList<String> names = new ArrayList<String>();
		ArrayList<Double> stdDevsLocal = new ArrayList<Double>();
		ArrayList<Integer> windows = new ArrayList<Integer>();

		for (Object o : anom.keySet()){
			String name = (String)o;
			if (name.equals("names")){
				namesA = anom.getProperty(name).split(";");
				for (String n : namesA){
					names.add(n);
				}
			} else if (name.equals("stdDevs")){
				for (String n : anom.getProperty(name).split(";")){
					stdDevsLocal.add(Double.parseDouble(n));
				}
			}else if (name.equals("windows")){
				for (String n : anom.getProperty(name).split(";")){
					windows.add(Integer.parseInt(n));
				}
			}
			else if (name.equals("directions")){
				directionsA = anom.getProperty(name).split(";");
			}
		}
		windowSizes.addAll(windows);
		stdDevs.addAll(stdDevsLocal);

		Double stdDevsA[] = stdDevs.toArray(new Double[]{});			
		if (namesA!=null && directionsA!=null){
			for (int i = 0; i < namesA.length; i++){
				String n = namesA[i];
				int direction = Integer.parseInt(directionsA[i]);
				for (int w : windows){
					MovingAverageConfig mac = new MovingAverageConfig();
					mac.windowSize = w;
					mac.anomalyDirection = direction;
					mac.stdDevThresh = stdDevsA;
					configs.put(n, mac); // TODO this is redunant over w's
				}
			}
		}
			
//			String value = (String)anom.get(o);
//			if (value!=null){
//				String parts[] = value.split("\\;");
//				if (parts.length==3){
//					MovingAverageConfig mac = new MovingAverageConfig();
//					mac.windowSize = Integer.parseInt(parts[0]);
//					windowSizes.add(mac.windowSize);
//					double sdMax = Double.parseDouble(parts[1]);
//					mac.anomalyDirection = Integer.parseInt(parts[2]);
//					double sdMin = Double.parseDouble(parts[3]);
//					mac.stdDevThresh = new Double[1+(int)((sdMax-sdMin)/stdDevIncr)];
//					for (int i = 0; i < mac.stdDevThresh.length; i++){
//						mac.stdDevThresh[i] = sdMin;
//						stdDevs.add(sdMin);
//						sdMin+=stdDevIncr;						
//					}
//					configs.put(name, mac);
//				}
//			}
//		}
		
	}
	
	public void writeProperties(ArrayList<String> locations){
		
		// output detection settings
		Properties props = new Properties();
		/*for (Entry<String, MovingAverageConfig> ent : configs.entrySet()){
			props.put(ent.getKey(), ent.getValue().toString());
		}*/
		props.put("names", Util.join(";", namesA));
		props.put("stdDevs", Util.join(";", stdDevs.toArray()));
		props.put("windows", Util.join(";", windowSizes.toArray()));
		props.put("directions", Util.join(";", directionsA));
		
		for (String filename : locations){
			filename = filename.replace("news.properties", AnomalyProperties.ANOMALY_DETECTION_PROPERTIES);
			try {
				FileOutputStream fos = new FileOutputStream(filename);
				props.store(fos, "");
				break;
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		// output detection history
		synchronized (histories){
			if (histories.size()>0){
				props = new Properties();
				int index = 0;
				for (ArrayList<MovingAverageHistory> mahs : histories.values()){
					for (MovingAverageHistory mah : mahs){
						mah.outputToString(props, index);
						index++;
					}
						
				}
				for (String filename : locations){
					filename = filename.replace("news.properties", ANOMALY_HISTORY_PROPERTIES);
					try {
						FileOutputStream fos = new FileOutputStream(filename);
						props.store(fos, "");
						break;
					} catch (FileNotFoundException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}
		
		}
		
	}
	
	public void changeStdDevs(double newVal){
		// TODO re-implement for new model
		throw new RuntimeException("Not implemented!");
//		synchronized (configs){
//			for (MovingAverageConfig conf : configs.values()){
//				conf.stdDevThresh=newVal;
//			}
//		}
	}
	
	public int getNumDataSources(){
		return configs.size();
	}


	public void loadAnomalyHistory(ArrayList<String> propertiesLocations) {
		synchronized (histories){
			Properties props = new Properties();
			for (String path : propertiesLocations){
				String newName = path.replace("news.properties", ANOMALY_HISTORY_PROPERTIES);
				File anom_props = new File(newName);
				if (anom_props.exists()){
					FileInputStream fis;
					try {
						fis = new FileInputStream(anom_props);
					
						props.load(fis);
					        				
						histories = MovingAverageHistory.loadFromProperties(props);
					
					break;
					} catch (FileNotFoundException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					}
				}
			}
		}
		
	}

	/**
	 * Returns the moving average histories for the network 
	 * containing the user.
	 * @return
	 */
	public ArrayList<MovingAverageHistory> getMovingAverageHistories() {
		NetworkAdminASNImpl asnInfo = NetworkAdminASNLookupImpl.getInstance().getASNInfo();
		String keys[] = new String[]{"Prefix:"+asnInfo.getBGPPrefix(), "AS:"+asnInfo.getAS(), "Country:"+asnInfo.getCountry()};
		ArrayList<MovingAverageHistory> toReturn = new ArrayList<MovingAverageHistory>();

		synchronized (histories){
			// if no history, create it now... will need window size and 
			// std dev settings from somewhere else...
			
			for (String key : keys){
				createMovingAverageHistory(key);
			}
			
			for (String key : keys){
				toReturn.addAll(histories.get(key));
			}
		}
		
		return toReturn;
		
	}

	public HashSet<Double> getStdDevs(String name) {	
		return stdDevs;
	}

	public HashSet<Integer> getWindowSizes() {
		return windowSizes;
	}

	public ArrayList<MovingAverageHistory> createMovingAverageHistory(String networkName) {
		synchronized (histories){
			if (histories.get(networkName)!=null) return histories.get(networkName);
			histories.put(networkName, new ArrayList<MovingAverageHistory>());
			for (int w : windowSizes){
				for (double d : stdDevs){
					MovingAverageHistory mah = new MovingAverageHistory();						
					mah.history = new BitSet(0);
					mah.stdDevThresh = d;
					mah.windowSize = w;
					mah.currentSession = new ArrayList<Boolean>();
					mah.networkName = networkName;						
					histories.get(networkName).add(mah);
				}
			}
			return histories.get(networkName);
		}
	}

}
