/**
 * NEWS Project
 *
 * File:         MovingAverage.java
 * RCS:          $Id: MovingAverage.java,v 1.11 2009/08/10 12:56:53 drc915 Exp $
 * Description:  MovingAverage class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 7, 2008 at 9:52:28 AM
 * Language:     Java
 * Package:      edu.northwestern.news.anomaly
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.news.anomaly;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;

import edu.northwestern.news.anomaly.AnomalyProperties.MovingAverageConfig;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The MovingAverage class ...
 */
public class MovingAverage {
	
	private static final boolean DEBUG = false;
	/** size of the moving average window */
	private int windowSize;
	/** threshold for anomaly detection */
	private HashSet<Double> stdDevThreshes;
	/** whether signal is in anomalous state */
	private HashMap<Double, Boolean> isAnomalous;
	/** values that are sampled */
	private ArrayList<Double> samples;
	/** values that are sampled */
	private ArrayList<Double> samplesAnomalous;
	/** mean */
	private double mean;
	/** variance */
	private double variance;
	/** sign of the change in value that is anomalous, 0 for any direction */
	private int anomalyDirection;
	private double deviationPct;
	private double anomalousStdDev = 0;

	
	public MovingAverage(int windowSize, double stdDevThresh[], 
			int anomalyDirection ){
		this.windowSize = windowSize;
		this.stdDevThreshes = new HashSet<Double>();
		isAnomalous = new HashMap<Double, Boolean>();
		for ( double t : stdDevThresh ) {
			stdDevThreshes.add(t);
			isAnomalous.put(t, false);
		}
		samples = new ArrayList<Double>(windowSize);
		samplesAnomalous = new ArrayList<Double>(windowSize);
		mean = 0;
		variance = 0;
		this.anomalyDirection = anomalyDirection;
		
	}
	
	public MovingAverage(MovingAverageConfig config) {
		this.windowSize = config.windowSize;
		this.stdDevThreshes = new HashSet<Double>();
		isAnomalous = new HashMap<Double, Boolean>();
		for ( double t : config.stdDevThresh ) {
			stdDevThreshes.add(t);
			isAnomalous.put(t, false);
		}
		samples = new ArrayList<Double>(windowSize);
		samplesAnomalous = new ArrayList<Double>(windowSize);
		mean = 0;
		variance = 0;
		this.anomalyDirection = config.anomalyDirection;
	}

	public void addSample(double value){
		if (DEBUG) System.out.println("Number of samples: "+samples.size());
		if (DEBUG && samples.size()==windowSize) System.out.println("Value: "+value+
				" Mean: "+mean + " StdDev:"+Math.sqrt(variance));
		
		for (double stdDevThresh : stdDevThreshes){
		if (samples.size()==windowSize){
		
			
			if (Math.abs(value-mean) > stdDevThresh*Math.sqrt(variance)){
				
				// only signal anomaly if in proper direction
				if (anomalyDirection!=0 && 
						Math.signum(value-mean)==anomalyDirection){
					// not only variation, but threshold drop in percent
					deviationPct = Math.abs(value-mean)/mean;
					isAnomalous.put(stdDevThresh, true);
					anomalousStdDev  = Math.abs(value-mean)/Math.sqrt(variance);
					if (DEBUG) System.out.println("Anomalous data!");
				}
				
				samplesAnomalous.add(value);
				if (samplesAnomalous.size()>windowSize/2){
					samples.clear();
					samples.addAll(samplesAnomalous);
					samplesAnomalous.clear();
				}
				
				return;
			} else {
				isAnomalous.put(stdDevThresh, false);
			}
			
		} else isAnomalous.put(stdDevThresh, false);
		} // end for each thresh
		
		if (samples.size()==windowSize) samples.remove(0);
		samples.add(value);

		updateStats();
	}
	
	private void updateStats(){
		mean = 0;
		for (double d : samples) mean+=d;
		mean/= samples.size();
		
		variance =0;
		for (double d : samples) variance += (d-mean)*(d-mean);
		variance/=samples.size();
		
	}
	
	public boolean isAnomalous(double stdDev){
		return isAnomalous.containsKey(stdDev) && isAnomalous.get(stdDev);
	}

	public int getWindowSize() {
		return windowSize;
	}

//	public double getStdDevThresh() {
//		return stdDevThresh;
//	}

	public boolean configEquals(MovingAverageConfig config) {
		if (config.anomalyDirection==anomalyDirection && 				
				config.windowSize== windowSize){
			if (config.stdDevThresh.length == this.stdDevThreshes.size()){
				for (double t : config.stdDevThresh) if (!stdDevThreshes.contains(t)) return false;
				return true;
			}			
		}
		return false;
	}

	public void removeSample() {
		if (samples.size()>0) {
			samples.remove(0);		
		}
		for (double k : isAnomalous.keySet()) isAnomalous.put(k, false); // because we don't want to leave it
							// in an anomalous state when no data is coming
	}

	public boolean isValid() {		
		return samples.size()>=windowSize;
	}

	public int getEmptyEntryCount() {
		return windowSize-samples.size();
	}
	
	public int entryCount(){
		return samples.size();
	}

	public double getDeviationPercent() {
		return deviationPct;
	}

	public double getAnomalousStdDev() {	
		return anomalousStdDev;
	}
}
